local f = CreateFrame("Frame")
f:Hide()

local update_icons_frame = CreateFrame("Frame")
update_icons_frame:Hide()

local pairs = pairs
local next = next
local type = type
local tonumber = tonumber
local string_upper = string.upper
local hooksecurefunc = hooksecurefunc
local GetActionInfo = GetActionInfo
local GetMacroIndexByName = GetMacroIndexByName
local GetMacroInfo = GetMacroInfo
local SetMacroSpell_original = SetMacroSpell
local SetMacroItem_original = SetMacroItem

local hook_all_actionbuttons
local SetTexture_hook
local CreateFrame_hook

local action_buttons = {}
local macros_override_icons = {}

local buttons_hooked
local hook_needed
local last_fix_needed
local hook_safe = true
local collection_blocked

function SetMacroIcon(macro_name_or_index,icon_file)
	macro_name_or_index = tonumber(macro_name_or_index) or macro_name_or_index
	local macro_name
	if type(macro_name_or_index)=="string" then
		macro_name = GetMacroIndexByName(macro_name_or_index)>0 and macro_name_or_index
	else
		macro_name = GetMacroInfo(macro_name_or_index)
	end
	if macro_name and icon_file~=macros_override_icons[macro_name] then
		macros_override_icons[macro_name] = icon_file
		update_icons_frame:Show()
	end
end

function GetMacroIcon(macro_name_or_index)
	macro_name_or_index = tonumber(macro_name_or_index) or macro_name_or_index
	local macro_name
	if type(macro_name_or_index)=="string" then
		macro_name = GetMacroIndexByName(macro_name_or_index)>0 and macro_name_or_index
	else
		macro_name = GetMacroInfo(macro_name_or_index)
	end
	return macros_override_icons[macro_name]
end

function SetMacroSpell(macro_name_or_index,spell_name,target)
	SetMacroIcon(macro_name_or_index)
	return SetMacroSpell_original(macro_name_or_index,spell_name,target)
end

function SetMacroItem(macro_name_or_index,item_name,target)
	SetMacroIcon(macro_name_or_index)
	return SetMacroItem_original(macro_name_or_index,item_name,target)
end

function SetTexture_hook(self,texture)
	if hook_needed and hook_safe then
		hook_safe = false
		local button,button_type,macro_index,override_icon
		button = self:GetParent()
		if texture then
			action_buttons[button].original_texture = texture
		elseif not action_buttons[button].original_texture then
			action_buttons[button].original_texture = self:GetTexture()
		end
		button_type = button:GetAttribute("type")
		if button_type=="action" then
			local action_index = button:GetAttribute("action") or button.action
			if action_index then
				local action_type,id = GetActionInfo(action_index)
				if action_type=="macro" then
					macro_index = id
				end
			end
		elseif button_type=="macro" then
			macro_index = button:GetAttribute("macro")
		end
		override_icon = macros_override_icons[GetMacroInfo(macro_index)]
		if override_icon then
			self:SetTexture(override_icon)
			button.Border:SetAlpha(0)
			button.Count:SetAlpha(0)
		else
			self:SetTexture(action_buttons[button].original_texture)
			button.Count:SetAlpha(1)
		end
		hook_safe = true
	end
end

function CreateFrame_hook(frame_type,_,_,frame_template)
	if string_upper(frame_type)=="CHECKBUTTON" and frame_template and string_upper(frame_template):match("ACTIONBUTTONTEMPLATE") then
		f:Show()
	end
end

function hook_all_actionbuttons(object)
	if not object:IsForbidden() then
		local object_type = object:GetObjectType()
		if object_type=="Frame" then
			for k,v in pairs({object:GetChildren()}) do
				hook_all_actionbuttons(v)
			end
		elseif object_type=="CheckButton" and object.GetAttribute and object.icon and object.Border and object.Count then
			if not action_buttons[object] then
				action_buttons[object] = {}
				hooksecurefunc(object.icon,"SetTexture",SetTexture_hook)
			end
		end
	end
end

update_icons_frame:SetScript("OnUpdate",function(self,elapsed)
	update_icons_frame:Hide()
	hook_needed = next(macros_override_icons) and true or false
	last_fix_needed = hook_needed or last_fix_needed
	if hook_needed then
		if not buttons_hooked then
			hook_all_actionbuttons(UIParent)
			if not collection_blocked then
				collectgarbage()
			else
				f:RegisterEvent("PLAYER_REGEN_ENABLED")
			end
			hooksecurefunc("CreateFrame",CreateFrame_hook)
			buttons_hooked = true
		end
		for k,v in pairs(action_buttons) do
			SetTexture_hook(k.icon)
		end
	elseif last_fix_needed then
		for k,v in pairs(action_buttons) do
			k.icon:SetTexture(v.original_texture)
			k.Count:SetAlpha(1)
			v.original_texture = nil
		end
		last_fix_needed = false
	end
end)

f:SetScript("OnUpdate",function(self,elapsed)
	self.timer = self.timer and self.timer+elapsed or 0
	if self.timer>1 then
		self.timer = 0
		self:Hide()
		hook_all_actionbuttons(UIParent)
		if not collection_blocked then
			collectgarbage()
		else
			self:RegisterEvent("PLAYER_REGEN_ENABLED")
		end
	end
end)

f:SetScript("OnEvent",function(self,event,...)
	if event=="UPDATE_MACROS" then
		for k,v in pairs(macros_override_icons) do
			if GetMacroIndexByName(k)==0 then
				macros_override_icons[k] = nil
			end
		end
	elseif event=="PLAYER_REGEN_DISABLED" then
		collection_blocked = true
	elseif event=="PLAYER_REGEN_ENABLED" then
		collection_blocked = false
		collectgarbage()
		self:UnregisterEvent("PLAYER_REGEN_ENABLED")
	end
end)
f:RegisterEvent("UPDATE_MACROS")
f:RegisterEvent("PLAYER_REGEN_DISABLED")